/*****************************************************************************
 * Copyright (C) 2008-2014
 *   
 * Michiel Broek <mbroek at mbse dot eu>
 *
 * This file is part of the mbsePi-apps
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * mbsePi-apps is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with EC-65K; see the file COPYING.  If not, write to the Free
 * Software Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *****************************************************************************/


#include "../lib/mbselib.h"
#include "server.h"
#include "thermferm.h"

extern bool		my_shutdown;
extern bool		debug;
extern int		lcdHandle;
extern unsigned char	lcdbuf[MAX_LCDS][20][4];
extern sys_config       Config;

int			s;		/* connected socket			*/
int			ls;		/* listen socket			*/

struct sockaddr_in	myaddr_in;	/* for local socket address             */
struct sockaddr_in	peeraddr_in;	/* for peer socket address              */

struct hostent		*hp;

#define SS_BUFSIZE      1024
#define SS_TIMEOUT      300

unsigned char		cs_mode = 'o';	/* o = Off, f = fridge, b = beer, p = profile-run */
float			cs_beerSet = 20.0;
float			cs_fridgeSet = 20.0;
float			cs_heatEstimator = 0.2;
float			cs_coolEstimator = 5;

unsigned char		cc_tempFormat = 'C';
float			cc_tempSetMin = 1.0;
float			cc_tempSetMax = 30.0;
float			cc_idleRangeH = 1.000;
float			cc_idleRangeL = -1.000;

float			cv_beerDiff = 0.0;



void defaultControlsettings(void)
{
	cs_mode = 'o';		/* o = Off, f = fridge, b = beer, p = profile-run */
	cs_beerSet = 20.0;
	cs_fridgeSet = 20.0;
	cs_heatEstimator = 0.2;
	cs_coolEstimator = 5;
}



void defaultControlConstants(void)
{
	cc_tempFormat = 'C';
	cc_tempSetMin = 1.0;
	cc_tempSetMax = 30.0;
	cc_idleRangeH = 1.000;
	cc_idleRangeL = -1.000;
}



void defaultControlVariables(void)
{
    cv_beerDiff = 0.0;
}



/*
 * Send message to client
 */
int srv_send(const char *format, ...)
{
    char        out[SS_BUFSIZE];
    va_list     va_ptr;

    if (s == -1)
	return -1;

    va_start(va_ptr, format);
    vsnprintf(out, SS_BUFSIZE-1, format, va_ptr);
    va_end(va_ptr);

    if (debug) {
    	syslog(LOG_NOTICE, "send: \"%s\"", out);
	fprintf(stdout, "send: \"%s\"\n", out);
    }

    if (send(s, out, strlen(out), 0) != strlen(out)) {
	syslog(LOG_NOTICE, "srv_send failed");
	return -1;
    }

    if (send(s, (char *)"\r\n", 2, 0) != 2) {
	syslog(LOG_NOTICE, "srv_send failed");
	return -1;
    }

    return 0;
}



void cmd_die(int onsig)
{
    syslog(LOG_NOTICE, "Server process die on signal %d", onsig);
    close(s);
    exit(0);
}



void cmd_server(void)
{
    char                *inp, *hostname, buf[SS_BUFSIZE], obuf[SS_BUFSIZE];
    int                 i, rc, rlen;
    socklen_t           fromlen;
    float		newtemp;

    hp = gethostbyaddr ((char *) &peeraddr_in.sin_addr, sizeof(struct in_addr), peeraddr_in.sin_family);
    if (hp == NULL) {
	hostname = inet_ntoa(peeraddr_in.sin_addr);
    } else {
	hostname = hp->h_name;
    }

    if (debug) {
    	syslog(LOG_NOTICE, "Start new client connection from %s port %u", hostname, ntohs(peeraddr_in.sin_port));
	fprintf(stdout, "Start new client connection from %s port %u\n", hostname, ntohs(peeraddr_in.sin_port));
    }

    memset((char *)&buf, 0, SS_BUFSIZE);
    fromlen = sizeof(peeraddr_in);
    rlen = recvfrom(s, buf, sizeof(buf) -1, 0, (struct sockaddr *)&peeraddr_in, &fromlen);
    if (rlen == -1) {
	syslog(LOG_NOTICE, "recvfrom(): %s", strerror(errno));
    } else {		    
	for (i = 0; i < strlen(buf); i++) {
	    if (buf[i] == '\n')
		buf[i] = '\0';
	    if (buf[i] == '\r')
	    	buf[i] = '\0';
	}
	if (strlen(buf)) {
	    if (debug) {
		syslog(LOG_NOTICE, "recv: \"%s\"", buf);
		fprintf(stdout, "recv: \"%s\"\n", buf);
	    }

	    /*
	     * Process commands from the client
	     */
	    if (strncmp(buf, "ack", 3) == 0) {
		srv_send((char *)"ack");
	    } else if (strncmp(buf, "lcd", 3) == 0) {
		sprintf(obuf, "[\"                    \", \"                    \", \"                    \", \"                    \"]");
		for (i = 0; i < 20; i++) {
		    obuf[i+2]  = lcdbuf[lcdHandle][i][0];
		    obuf[i+26] = lcdbuf[lcdHandle][i][1];
		    obuf[i+50] = lcdbuf[lcdHandle][i][2];
		    obuf[i+74] = lcdbuf[lcdHandle][i][3];
		}
		srv_send(obuf);
	    } else if (strncmp(buf, "getMode", 7) == 0) {
		srv_send("%c", cs_mode);
	    } else if (strncmp(buf, "getFridge", 9) == 0) {
		srv_send("%.1f", cs_fridgeSet);
	    } else if (strncmp(buf, "getBeer", 7) == 0) {
		srv_send("%.1f", cs_beerSet);
	    } else if (strncmp(buf, "getControlConstants", 19) == 0) {
		srv_send("{ \"tempFormat\":\"%c\", \"tempSetMin\":%.1f, \"tempSetMax\":%.1f, \"idleRangeH\":%.3f, \"idleRangeL\":%.3f }", 
			cc_tempFormat, cc_tempSetMin, cc_tempSetMax, cc_idleRangeH, cc_idleRangeL );
	    } else if (strncmp(buf, "getControlSettings", 18) == 0) {
		srv_send("{ \"mode\":\"%c\", \"beerSet\":%.1f, \"fridgeSet\":%.1f, \"heatEstimator\":%.1f, \"coolEstimator\":%.1f }", 
			cs_mode, cs_beerSet, cs_fridgeSet, cs_heatEstimator, cs_coolEstimator);
	    } else if (strncmp(buf, "getControlVariables", 19) == 0) {
		srv_send("{ \"beerDiff\":%.2f }", cv_beerDiff);
	    } else if (strncmp(buf, "setBeer=", 8) == 0) {
		inp = xstrcpy(buf+8);
		rc = sscanf(inp, "%f", &newtemp);
		if (debug)
		    fprintf(stdout, "new temp from %s, %.1f, rc=%d\n", inp, newtemp, rc);
		if (rc == 1) {
		    srv_send("ack");
		    cs_mode = 'b';
		    cs_beerSet = newtemp;
		} else {
		    srv_send("err");
		}
		free(inp);
	    } else if (strncmp(buf, "setFridge=", 10) == 0) {
		inp = xstrcpy(buf+10);
		rc = sscanf(inp, "%f", &newtemp);
		if (debug)
		    fprintf(stdout, "new temp from %s, %.1f, rc=%d\n", inp, newtemp, rc);
		if (rc == 1) {
		    srv_send("ack");
		    cs_mode = 'f';
		    cs_fridgeSet = newtemp;
		} else {
		    srv_send("err");
		}
		free(inp);
	    } else {
		if (debug)
		    fprintf(stdout, "unknown command \"%s\"\n", buf);
		srv_send((char *)"ERR");
	    }
	}
    }

    if (debug) {
    	syslog(LOG_NOTICE, "End connection from %s port %u", hostname, ntohs(peeraddr_in.sin_port));
	fprintf(stdout, "End connection from %s port %u\n", hostname, ntohs(peeraddr_in.sin_port));
    }

    close(s);
}



PI_THREAD (my_server_loop)
{
    socklen_t   addrlen;
    int         optval = 1;

    syslog(LOG_NOTICE, "Thread my_server_loop started");
    if (debug)
	fprintf(stdout, "Thread my_server_loop started\n");

    memset((char *)&myaddr_in, 0, sizeof(struct sockaddr_in));
    memset((char *)&peeraddr_in, 0, sizeof(struct sockaddr_in));
    myaddr_in.sin_family = AF_INET;
    myaddr_in.sin_addr.s_addr = INADDR_ANY;
    myaddr_in.sin_port = htons(Config.my_port);

    ls = socket(AF_INET, SOCK_STREAM, 0);
    if (ls == -1) {
	syslog(LOG_NOTICE, "Can't create listen socket: %s", strerror(errno));
	fprintf(stderr, "Can't create listen socket: %s\n", strerror(errno));
	return 0;
    }

    if (setsockopt(ls, SOL_SOCKET, SO_KEEPALIVE, &optval, sizeof(optval)) == -1) {
	syslog(LOG_NOTICE, "Can't setsockopt SO_KEEPALIVE socket: %s", strerror(errno));
	close(ls);
	return 0;
    }

    if (setsockopt(ls, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval)) == -1) {
	syslog(LOG_NOTICE, "Can't setsockopt SO_REUSEADDR socket: %s", strerror(errno));
	close(ls);
	return NULL;
    }

    if (bind(ls, (struct sockaddr *)&myaddr_in, sizeof(struct sockaddr_in)) == -1) {
	syslog(LOG_NOTICE, "Can't bind to listen socket: %s", strerror(errno));
	close(ls);
	return NULL;
    }

    if (listen(ls, 5) == -1) {
	syslog(LOG_NOTICE, "Can't listen on listen socket: %s", strerror(errno));
	close(ls);
	return NULL;
    }

    syslog(LOG_NOTICE, "listen socket created %d", ls);
    if (debug)
	fprintf(stdout, "listen socket created %d\n", ls);


    /*
     * Loop forever until the external shutdown variable is set.
     */
    for (;;) {
	
	addrlen = sizeof(struct sockaddr_in);
        /*
	 * This call will block until a new connection
	 * arrives. Then it will return the address of
	 * the connecting peer, and a new socket
	 * descriptor, s, for that connection.
	 */
	s = accept(ls, (struct sockaddr *)&peeraddr_in, &addrlen);
	if (s == -1) {
	    syslog(LOG_NOTICE, "my_server_loop accept failed %s", strerror(errno));
	    if (debug)
		fprintf(stdout, "my_server_loop accept failed %s\n", strerror(errno));
	    return 0;
	}

	cmd_server();

	if (my_shutdown) {
	    syslog(LOG_NOTICE, "Thread my_server_loop stopped");
	    if (debug)
		fprintf(stdout, "Thread my_server_loop stopped\n");
	    return 0;
	}

    }
}



